from collections import OrderedDict
from enum import IntEnum
from functools import partialmethod, lru_cache

from spherov2.commands.api_and_shell import ApiAndShell
from spherov2.commands.connection import Connection
from spherov2.commands.drive import Drive
from spherov2.commands.firmware import Firmware
from spherov2.commands.io import IO
from spherov2.commands.power import Power
from spherov2.commands.sensor import Sensor
from spherov2.commands.system_info import SystemInfo
from spherov2.controls.v2 import AnimationControl, DriveControl, LedControl, SensorControl, StatsControl, Processors
from spherov2.toy import ToyV2, Toy, ToySensor
from spherov2.types import ToyType


class BOLT(ToyV2):
    toy_type = ToyType('Sphero BOLT', 'SB-', 'SB', .075)
    _handshake = [('00020005-574f-4f20-5370-6865726f2121', bytearray(b'usetheforce...band'))] #TODO: find correct info here

    class LEDs(IntEnum):
        FRONT_RED = 0
        FRONT_GREEN = 1
        FRONT_BLUE = 2
        BACK_RED = 3
        BACK_GREEN = 4
        BACK_BLUE = 5

    # Sensors Available:
    sensors = OrderedDict(
        quaternion=OrderedDict(
            x=ToySensor(0x2000000, -1., 1.),
            y=ToySensor(0x1000000, -1., 1.),
            z=ToySensor(0x800000, -1., 1.),
            w=ToySensor(0x400000, -1., 1.)
        ),
        attitude=OrderedDict(
            pitch=ToySensor(0x40000, -179., 180.),
            roll=ToySensor(0x20000, -179., 180.),
            yaw=ToySensor(0x10000, -179., 180.)
        ),
        accelerometer=OrderedDict(
            x=ToySensor(0x8000, -8.19, 8.19),
            y=ToySensor(0x4000, -8.19, 8.19),
            z=ToySensor(0x2000, -8.19, 8.19)
        ),
        accel_one=OrderedDict(accel_one=ToySensor(0x200, 0., 8000.)),
        locator=OrderedDict(
            x=ToySensor(0x40, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x20, -32768., 32767., lambda x: x * 100.),
        ),
        velocity=OrderedDict(
            x=ToySensor(0x10, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x8, -32768., 32767., lambda x: x * 100.),
        ),
        speed=OrderedDict(speed=ToySensor(0x4, 0., 32767.)),  # Doesn't work when enabled
        core_time=OrderedDict(core_time=ToySensor(0x2, 0., 0.))
    )

    extended_sensors = OrderedDict(
        gyroscope=OrderedDict(
            x=ToySensor(0x2000000, -20000., 20000.),
            y=ToySensor(0x1000000, -20000., 20000.),
            z=ToySensor(0x800000, -20000., 20000.)
        ),
        ambient_light=OrderedDict(
            ambient_light=ToySensor(0x40000, 120000., 120000.))
    )

    # Bolt Supported calls
    # API and Shell:
    get_api_protocol_version = ApiAndShell.get_api_protocol_version  # GetApiProtocolVersionCommand
    get_supported_cids = ApiAndShell.get_supported_cids  # GetSupportedCidsCommand
    get_supported_dids = ApiAndShell.get_supported_dids  # GetSupportedDidsCommand
    ping = ApiAndShell.ping  # PingCommand
    send_command_to_shell = ApiAndShell.send_command_to_shell  # SendCommandToShellCommand
    add_send_string_to_console_listener = partialmethod(Toy._add_listener,
                                                        ApiAndShell.send_string_to_console)  # SendStringToConsoleCommand
    remove_send_string_to_console_listener = partialmethod(Toy._remove_listener,
                                                           ApiAndShell.send_string_to_console)  # SendStringToConsoleCommand

    # Connection
    get_bluetooth_advertising_name = Connection.get_bluetooth_advertising_name  # GetBluetoothAdvertisingNameCommand
    get_bluetooth_name = Connection.get_bluetooth_name  # GetBluetoothNameCommand
    set_bluetooth_name = Connection.set_bluetooth_name  # SetBluetoothNameCommand

    #Drive - Driving done on the Secondary Processor
    drive_with_heading = partialmethod(Drive.drive_with_heading,
                                                      proc=Processors.SECONDARY) #DriveWithHeadingCommand
    reset_yaw = partialmethod(Drive.reset_yaw,
                                                      proc=Processors.SECONDARY) #ResetYawCommand
    set_pitch_torque_modification_value = partialmethod(Drive.set_pitch_torque_modification_value,
                                                      proc=Processors.SECONDARY) #SetPitchTorqueModificationValueCommand
    set_raw_motors = partialmethod(Drive.set_raw_motors,
                                                      proc=Processors.SECONDARY) #SetRawMotorsCommand
    set_stabilization = partialmethod(Drive.set_stabilization,
                                                      proc=Processors.SECONDARY) #SetStabilizationCommand

    # Firmware
    get_pending_update_flags = Firmware.get_pending_update_flags  # GetPendingUpdateFlagsCommand

    # IO - most LED/Matrix stuff is done on Secondary Processor, if issues check that
    assign_compressed_frame_player_frames_to_animation = partialmethod(
        IO.assign_compressed_frame_player_frames_to_animation,
        proc=Processors.SECONDARY)  # AssignCompressedFramePlayerFramesToAnimationCommand
    add_compressed_frame_player_animation_complete_notify_listener = partialmethod(
        Toy._add_listener,
        IO.compressed_frame_player_animation_complete_notify)  # CompressedFramePlayerAnimationCompleteNotifyCommand
    remove_compressed_frame_player_animation_complete_notify_listener = partialmethod(
        Toy._remove_listener,
        IO.compressed_frame_player_animation_complete_notify)  # CompressedFramePlayerAnimationCompleteNotifyCommand
    delete_all_compressed_frame_player_animations_and_frames = partialmethod(
        IO.delete_all_compressed_frame_player_animations_and_frames,
        proc=Processors.SECONDARY)  # DeleteAllCompressedFramePlayerAnimationsAndFramesCommand
    draw_compressed_frame_player_fill = partialmethod(IO.draw_compressed_frame_player_fill,
                                                      proc=Processors.SECONDARY)  # DrawCompressedFramePlayerFillCommand
    draw_compressed_frame_player_line = partialmethod(IO.draw_compressed_frame_player_line,
                                                      proc=Processors.SECONDARY)  # DrawCompressedFramePlayerLineCommand
    get_compressed_frame_player_list_of_frames = partialmethod(IO.get_compressed_frame_player_list_of_frames,
                                                               proc=Processors.SECONDARY)  # GetCompressedFramePlayerListOfFramesCommand
    override_compressed_frame_player_animation_global_settings = partialmethod(
        IO.override_compressed_frame_player_animation_global_settings,
        proc=Processors.SECONDARY)  # OverrideCompressedFramePlayerAnimationGlobalSettingsCommand
    pause_compressed_frame_player_animation = partialmethod(IO.pause_compressed_frame_player_animation,
                                                            proc=Processors.SECONDARY)  # PauseCompressedFramePlayerAnimationCommand
    play_compressed_frame_player_animation = partialmethod(IO.play_compressed_frame_player_animation,
                                                           proc=Processors.SECONDARY)  # PlayCompressedFramePlayerAnimationCommand
    play_compressed_frame_player_animation_with_loop_option = partialmethod(
        IO.play_compressed_frame_player_animation_with_loop_option,
        proc=Processors.SECONDARY)  # PlayCompressedFramePlayerAnimationWithLoopOptionCommand
    play_compressed_frame_player_frame = partialmethod(IO.play_compressed_frame_player_frame,
                                                       proc=Processors.SECONDARY)  # PlayCompressedFramePlayerFrameCommand
    reset_compressed_frame_player_animation = partialmethod(IO.reset_compressed_frame_player_animation,
                                                            proc=Processors.SECONDARY)  # ResetCompressedFramePlayerAnimationCommand
    resume_compressed_frame_player_animation = partialmethod(IO.resume_compressed_frame_player_animation,
                                                             proc=Processors.SECONDARY)  # ResumeCompressedFramePlayerAnimationCommand
    save_compressed_frame_player64_bit_frame = partialmethod(IO.save_compressed_frame_player64_bit_frame,
                                                             proc=Processors.SECONDARY)  # SaveCompressedFramePlayer64BitFrameCommand
    save_compressed_frame_player_animation = partialmethod(IO.save_compressed_frame_player_animation,
                                                           proc=Processors.SECONDARY)  # SaveCompressedFramePlayerAnimationCommand
    save_compressed_frame_player_animation_without_frames = partialmethod(
        IO.save_compressed_frame_player_animation_without_frames,
        proc=Processors.SECONDARY)  # SaveCompressedFramePlayerAnimationWithoutFramesCommand
    set_all_leds_with_8_bit_mask = partialmethod(IO.set_all_leds_with_8_bit_mask,
                                                 proc=Processors.PRIMARY)  # SetAllLedsWith8BitMaskCommand
    set_compressed_frame_player = partialmethod(IO.set_compressed_frame_player,
                                                proc=Processors.SECONDARY)  # SetCompressedFramePlayerCommand
    set_compressed_frame_player_frame_rotation = partialmethod(IO.set_compressed_frame_player_frame_rotation,
                                                               proc=Processors.SECONDARY)  # SetCompressedFramePlayerFrameRotationCommand
    set_compressed_frame_player_one_color = partialmethod(IO.set_compressed_frame_player_one_color,
                                                          proc=Processors.SECONDARY)  # SetCompressedFramePlayerOneColorCommand
    set_compressed_frame_player_pixel = partialmethod(IO.set_compressed_frame_player_pixel,
                                                      proc=Processors.SECONDARY)  # SetCompressedFramePlayerPixelCommand
    set_compressed_frame_player_single_character = partialmethod(IO.set_compressed_frame_player_single_character,
                                                                 proc=Processors.SECONDARY)  # SetCompressedFramePlayerSingleCharacterCommand
    set_compressed_frame_player_text_scrolling = partialmethod(IO.set_compressed_frame_player_text_scrolling,
                                                               proc=Processors.SECONDARY)  # SetCompressedFramePlayerTextScrollingCommand
    add_set_compressed_frame_player_text_scrolling_notify = partialmethod(
        Toy._add_listener,
        IO.set_compressed_frame_player_text_scrolling_notify)  # SetCompressedFramePlayerTextScrollingNotifyCommand
    remove_set_compressed_frame_player_text_scrolling_notify = partialmethod(
        Toy._remove_listener,
        IO.set_compressed_frame_player_text_scrolling_notify)  # SetCompressedFramePlayerTextScrollingNotifyCommand
    set_led = partialmethod(IO.set_led, proc=Processors.SECONDARY)  # SetLedCommand

    # Testing
    release_led_requests = partialmethod(IO.release_led_requests, proc=Processors.PRIMARY)

    # Power
    add_charger_state_changed_notify_listener = partialmethod(Toy._add_listener,
                                                   Power.charger_state_changed_notify)  # ChargerStateChangedNotifyCommand
    remove_charger_state_changed_notify_listener = partialmethod(Toy._remove_listener,
                                                      Power.charger_state_changed_notify)  # ChargerStateChangedNotifyCommand
    add_did_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                   Power.did_sleep_notify)  # DidSleepNotifyCommand
    remove_did_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                      Power.did_sleep_notify)  # DidSleepNotifyCommand
    enable_battery_voltage_state_change_notify = Power.enable_battery_voltage_state_change_notify  # EnableBatteryVoltageStateChangeNotifyCommand
    enable_charger_state_changed_notify = Power.enable_charger_state_changed_notify  # EnableChargerStateChangedNotifyCommand
    enter_deep_sleep = Power.enter_deep_sleep  # EnterDeepSleepCommand
    get_battery_voltage = Power.get_battery_voltage  # GetBatteryVoltageCommand
    get_battery_voltage_state = Power.get_battery_voltage_state  # GetBatteryVoltageStateCommand
    get_charger_state = Power.get_charger_state  # GetChargerStateCommand
    sleep = Power.sleep  # SleepCommand
    wake = Power.wake  # WakeCommand
    add_will_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                   Power.will_sleep_notify)  # WillSleepNotifyCommand
    remove_will_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                      Power.will_sleep_notify)  # WillSleepNotifyCommand

    # Sensor
    add_collision_detected_notify_listener = partialmethod(Toy._add_listener,
                                                           Sensor.collision_detected_notify)  # CollisionDetectedNotifyCommand
    remove_collision_detected_notify_listener = partialmethod(Toy._remove_listener,
                                                              Sensor.collision_detected_notify)  # CollisionDetectedNotifyCommand
    configure_collision_detection = partialmethod(Sensor.configure_collision_detection, proc=Processors.SECONDARY)  # ConfigureCollisionDetectionCommand
    enable_gyro_max_notify = partialmethod(Sensor.enable_gyro_max_notify, proc=Processors.SECONDARY)  # EnableGyroMaxNotifyCommand
    get_ambient_light_sensor_value = partialmethod(Sensor.get_ambient_light_sensor_value, proc=Processors.SECONDARY)
    get_bot_to_bot_infrared_readings = Sensor.get_bot_to_bot_infrared_readings  # GetBotToBotInfraredReadingsCommand
    get_sensor_streaming_mask = partialmethod(Sensor.get_sensor_streaming_mask, proc=Processors.SECONDARY)  # GetSensorStreamingMaskCommand
    set_sensor_streaming_mask = partialmethod(Sensor.set_sensor_streaming_mask, proc=Processors.SECONDARY)  # SetSensorStreamingMaskCommand
    get_extended_sensor_streaming_mask = partialmethod(Sensor.get_extended_sensor_streaming_mask, proc=Processors.SECONDARY)  # GetExtendedSensorStreamingMaskCommand
    set_extended_sensor_streaming_mask = partialmethod(Sensor.set_extended_sensor_streaming_mask, proc=Processors.SECONDARY)  # SetExtendedSensorStreamingMaskCommand
    add_gyro_max_notify_listener = partialmethod(Toy._add_listener, Sensor.gyro_max_notify)  # GyroMaxNotifyCommand
    remove_gyro_max_notify_listener = partialmethod(Toy._remove_listener,
                                                    Sensor.gyro_max_notify)  # GyroMaxNotifyCommand
    listen_for_robot_to_robot_infrared_message = Sensor.listen_for_robot_to_robot_infrared_message  # ListenForRobotToRobotInfraredMessageCommand
    magnetometer_calibrate_to_north = partialmethod(Sensor.magnetometer_calibrate_to_north, proc=Processors.SECONDARY)  # MagnetometerCalibrateToNorthCommand
    add_magnetometer_north_yaw_notify_listener = partialmethod(Toy._add_listener,
                                                               Sensor.magnetometer_north_yaw_notify)  # MagnetometerNorthYawNotifyCommand
    remove_magnetometer_north_yaw_notify_listener = partialmethod(Toy._remove_listener,
                                                                  Sensor.magnetometer_north_yaw_notify)  # MagnetometerNorthYawNotifyCommand
    reset_locator_x_and_y = partialmethod(Sensor.reset_locator_x_and_y, proc=Processors.SECONDARY)  # ResetLocationXAndYCommand
    add_robot_to_robot_infrared_message_received_notify_listener = partialmethod(Toy._add_listener,
                                                                                 Sensor.robot_to_robot_infrared_message_received_notify)  ##RobotToRobotInfraredMessageReceivedNotifyCommand
    remove_robot_to_robot_infrared_message_received_notify_listener = partialmethod(Toy._remove_listener,
                                                                                    Sensor.robot_to_robot_infrared_message_received_notify)  # RobotToRobotInfraredMessageReceivedNotifyCommand
    send_robot_to_robot_infrared_message = Sensor.send_robot_to_robot_infrared_message  # SendRobotToRobotInfraredMessageCommand
    add_sensor_streaming_data_notify_listener = partialmethod(Toy._add_listener,
                                                              Sensor.sensor_streaming_data_notify)  # SensorStreamingDataNotifyCommand
    remove_sensor_streaming_data_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Sensor.sensor_streaming_data_notify)  # SensorStreamingDataNotifyCommand
    set_locator_flags = partialmethod(Sensor.set_locator_flags, proc=Processors.SECONDARY)  # SetLocatorFlagsCommand
    start_robot_to_robot_infrared_broadcasting = Sensor.start_robot_to_robot_infrared_broadcasting  # StartRobotToRobotInfraredBroadcastingCommand
    start_robot_to_robot_infrared_evading = Sensor.start_robot_to_robot_infrared_evading  # StartRobotToRobotInfraredEvadingCommand
    start_robot_to_robot_infrared_following = Sensor.start_robot_to_robot_infrared_following  # StartRobotToRobotInfraredFollowingCommand
    stop_robot_to_robot_infrared_broadcasting = Sensor.stop_robot_to_robot_infrared_broadcasting  # StopRobotToRobotInfraredBroadcastingCommand
    stop_robot_to_robot_infrared_evading = Sensor.stop_robot_to_robot_infrared_evading  # StopRobotToRobotInfraredEvadingCommand
    stop_robot_to_robot_infrared_following = Sensor.stop_robot_to_robot_infrared_following  # StopRobotToRobotInfraredFollowingCommand

    # System Info
    erase_config_block = SystemInfo.erase_config_block  # EraseConfigBlockCommand
    get_board_revision = SystemInfo.get_board_revision  # GetBoardRevisionCommand
    get_boot_reason = SystemInfo.get_boot_reason  # GetBootReasonCommand
    get_bootloader_version = SystemInfo.get_bootloader_version  # GetBootloaderVersionCommand
    get_secondary_mcu_bootloader_version = SystemInfo.get_secondary_mcu_bootloader_version
    get_config_block = SystemInfo.get_config_block  # GetConfigBlockCommand
    get_last_error_info = SystemInfo.get_last_error_info  # GetLastErrorInfoCommand
    get_mac_address = SystemInfo.get_mac_address  # GetMacAddressCommand
    get_main_app_version = SystemInfo.get_main_app_version  # GetMainAppVersionCommand
    get_secondary_main_app_version = SystemInfo.get_secondary_main_app_version
    get_manufacturing_date = SystemInfo.get_manufacturing_date  # GetManufacturingDateCommand
    get_processor_name = SystemInfo.get_processor_name  # GetProcessorNameCommand
    get_sku = SystemInfo.get_sku  # GetSkuCommand
    get_stats_id = SystemInfo.get_stats_id  # GetStatsIdCommand
    get_swd_locking_status = SystemInfo.get_swd_locking_status  # GetSwdLockingStatusCommand
    set_config_block = SystemInfo.set_config_block  # SetConfigBlockCommand
    write_config_block = SystemInfo.write_config_block  # WriteConfigBlockCommand

    # Controls - V2
    @property
    @lru_cache(None)
    def animation_control(self):
        return AnimationControl(self)

    @property
    @lru_cache(None)
    def drive_control(self):
        return DriveControl(self)

    @property
    @lru_cache(None)
    def multi_led_control(self):
        return LedControl(self)

    @property
    def sensor_control(self):
        if self._sensor_controller is None: #Make a new SensorControl if necessary (NOT A UTIL CLASS)
            self._sensor_controller = SensorControl(self)
        return self._sensor_controller

    @property
    @lru_cache(None)
    def stats_control(self):
        return StatsControl(self)



from enum import Enum


class RequestOp(bytes, Enum):
    SCAN = b'\x00'
    INIT = b'\x01'
    SET_CALLBACK = b'\x02'
    WRITE = b'\x03'
    FIND = b'\x04'
    END = b'\xff'


class ResponseOp(bytes, Enum):
    OK = b'\x00'
    ON_DATA = b'\x01'
    ERROR = b'\xff'



from collections import OrderedDict
from enum import IntEnum
from functools import partialmethod, lru_cache

from spherov2.commands.animatronic import Animatronic
from spherov2.commands.api_and_shell import ApiAndShell
from spherov2.commands.connection import Connection
from spherov2.commands.drive import Drive
from spherov2.commands.factory_test import FactoryTest
from spherov2.commands.firmware import Firmware
from spherov2.commands.io import IO
from spherov2.commands.power import Power
from spherov2.commands.sensor import Sensor
from spherov2.commands.system_info import SystemInfo
from spherov2.controls.v2 import DriveControl, LedControl, SensorControl, FirmwareUpdateControl, StatsControl
from spherov2.toy import ToyV2, ToySensor, Toy
from spherov2.types import ToyType


class BB9E(ToyV2):
    toy_type = ToyType('BB-9E', 'GB-', 'GB', .12)
    _handshake = [('00020005-574f-4f20-5370-6865726f2121', bytearray(b'usetheforce...band'))]

    class LEDs(IntEnum):
        BODY_RED = 0
        BODY_GREEN = 1
        BODY_BLUE = 2
        AIMING = 3
        HEAD = 4

    class Animations(IntEnum):
        EMOTE_ALARM = 0
        EMOTE_NO = 1
        EMOTE_SCAN_SWEEP = 2
        EMOTE_SCARED = 3
        EMOTE_YES = 4
        EMOTE_AFFIRMATIVE = 5
        EMOTE_AGITATED = 6
        EMOTE_ANGRY = 7
        EMOTE_CONTENT = 8
        EMOTE_EXCITED = 9
        EMOTE_FIERY = 10
        EMOTE_GREETINGS = 11
        EMOTE_NERVOUS = 12
        EMOTE_SLEEP = 14
        EMOTE_SURPRISED = 15
        EMOTE_UNDERSTOOD = 16
        HIT = 17
        WWM_ANGRY = 18
        WWM_ANXIOUS = 19
        WWM_BOW = 20
        WWM_CURIOUS = 22
        WWM_DOUBLE_TAKE = 23
        WWM_EXCITED = 24
        WWM_FIERY = 25
        WWM_HAPPY = 26
        WWM_JITTERY = 27
        WWM_LAUGH = 28
        WWM_LONG_SHAKE = 29
        WWM_NO = 30
        WWM_OMINOUS = 31
        WWM_RELIEVED = 32
        WWM_SAD = 33
        WWM_SCARED = 34
        WWM_SHAKE = 35
        WWM_SURPRISED = 36
        WWM_TAUNTING = 37
        WWM_WHISPER = 38
        WWM_YELLING = 39
        WWM_YOOHOO = 40
        WWM_FRUSTRATED = 41
        IDLE_1 = 42
        IDLE_2 = 43
        IDLE_3 = 44
        EYE_1 = 45
        EYE_2 = 46
        EYE_3 = 47
        EYE_4 = 48

    sensors = OrderedDict(
        quaternion=OrderedDict(
            x=ToySensor(0x2000000, -1., 1.),
            y=ToySensor(0x1000000, -1., 1.),
            z=ToySensor(0x800000, -1., 1.),
            w=ToySensor(0x400000, -1., 1.)
        ),
        attitude=OrderedDict(
            pitch=ToySensor(0x40000, -179., 180.),
            roll=ToySensor(0x20000, -179., 180.),
            yaw=ToySensor(0x10000, -179., 180.)
        ),
        accelerometer=OrderedDict(
            x=ToySensor(0x8000, -8.19, 8.19),
            y=ToySensor(0x4000, -8.19, 8.19),
            z=ToySensor(0x2000, -8.19, 8.19)
        ),
        accel_one=OrderedDict(accel_one=ToySensor(0x200, 0., 8000.)),
        locator=OrderedDict(
            x=ToySensor(0x40, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x20, -32768., 32767., lambda x: x * 100.),
        ),
        velocity=OrderedDict(
            x=ToySensor(0x10, -32768., 32767., lambda x: x * 100.),
            y=ToySensor(0x8, -32768., 32767., lambda x: x * 100.),
        ),
        speed=OrderedDict(speed=ToySensor(0x4, 0., 32767.)),
        core_time=OrderedDict(core_time=ToySensor(0x2, 0., 0.))
    )

    extended_sensors = OrderedDict(
        gyroscope=OrderedDict(
            x=ToySensor(0x2000000, -20000., 20000.),
            y=ToySensor(0x1000000, -20000., 20000.),
            z=ToySensor(0x800000, -20000., 20000.)
        )
    )

    # Animatronic
    play_animation = Animatronic.play_animation
    stop_animation = Animatronic.stop_animation
    enable_idle_animations = Animatronic.enable_idle_animations
    enable_trophy_mode = Animatronic.enable_trophy_mode
    get_trophy_mode_enabled = Animatronic.get_trophy_mode_enabled

    # APIAndShell
    ping = ApiAndShell.ping
    get_api_protocol_version = ApiAndShell.get_api_protocol_version
    send_command_to_shell = ApiAndShell.send_command_to_shell
    add_send_string_to_console_listener = partialmethod(Toy._add_listener, ApiAndShell.send_string_to_console)
    remove_send_string_to_console_listener = partialmethod(Toy._remove_listener, ApiAndShell.send_string_to_console)

    # Connection
    set_bluetooth_name = Connection.set_bluetooth_name
    get_bluetooth_name = Connection.get_bluetooth_name

    # Drive
    set_raw_motors = Drive.set_raw_motors
    reset_yaw = Drive.reset_yaw
    drive_with_heading = Drive.drive_with_heading
    set_stabilization = Drive.set_stabilization

    # FactoryTest
    get_factory_mode_challenge = FactoryTest.get_factory_mode_challenge
    enter_factory_mode = FactoryTest.enter_factory_mode
    exit_factory_mode = FactoryTest.exit_factory_mode
    get_chassis_id = FactoryTest.get_chassis_id

    # Firmware
    get_pending_update_flags = Firmware.get_pending_update_flags

    # IO
    play_audio_file = IO.play_audio_file
    set_all_leds_with_16_bit_mask = IO.set_all_leds_with_16_bit_mask
    start_idle_led_animation = IO.start_idle_led_animation

    # Sensor
    set_sensor_streaming_mask = Sensor.set_sensor_streaming_mask
    get_sensor_streaming_mask = Sensor.get_sensor_streaming_mask
    add_sensor_streaming_data_notify_listener = partialmethod(Toy._add_listener, Sensor.sensor_streaming_data_notify)
    remove_sensor_streaming_data_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Sensor.sensor_streaming_data_notify)
    set_extended_sensor_streaming_mask = Sensor.set_extended_sensor_streaming_mask
    get_extended_sensor_streaming_mask = Sensor.get_extended_sensor_streaming_mask
    enable_gyro_max_notify = Sensor.enable_gyro_max_notify
    add_gyro_max_notify_listener = partialmethod(Toy._add_listener, Sensor.gyro_max_notify)
    remove_gyro_max_notify_listener = partialmethod(Toy._remove_listener, Sensor.gyro_max_notify)
    configure_collision_detection = Sensor.configure_collision_detection
    add_collision_detected_notify_listener = partialmethod(Toy._add_listener, Sensor.collision_detected_notify)
    remove_collision_detected_notify_listener = partialmethod(Toy._remove_listener, Sensor.collision_detected_notify)
    reset_locator_x_and_y = Sensor.reset_locator_x_and_y
    set_locator_flags = Sensor.set_locator_flags
    set_accelerometer_activity_threshold = Sensor.set_accelerometer_activity_threshold
    enable_accelerometer_activity_notify = Sensor.enable_accelerometer_activity_notify
    add_accelerometer_activity_notify_listener = partialmethod(Toy._add_listener, Sensor.accelerometer_activity_notify)
    remove_accelerometer_activity_notify_listener = partialmethod(Toy._remove_listener,
                                                                  Sensor.accelerometer_activity_notify)
    set_gyro_activity_threshold = Sensor.set_gyro_activity_threshold
    enable_gyro_activity_notify = Sensor.enable_gyro_activity_notify
    add_gyro_activity_notify_listener = partialmethod(Toy._add_listener, Sensor.gyro_activity_notify)
    remove_gyro_activity_notify_listener = partialmethod(Toy._remove_listener, Sensor.gyro_activity_notify)

    # Power
    enter_deep_sleep = Power.enter_deep_sleep
    sleep = Power.sleep
    get_battery_voltage = Power.get_battery_voltage
    get_battery_state = Power.get_battery_state
    enable_battery_state_changed_notify = Power.enable_battery_state_changed_notify
    add_battery_state_changed_notify_listener = partialmethod(Toy._add_listener, Power.battery_state_changed_notify)
    remove_battery_state_changed_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Power.battery_state_changed_notify)
    wake = Power.wake
    add_will_sleep_notify_listener = partialmethod(Toy._add_listener, Power.will_sleep_notify)
    remove_will_sleep_notify_listener = partialmethod(Toy._remove_listener, Power.will_sleep_notify)
    add_did_sleep_notify_listener = partialmethod(Toy._add_listener, Power.did_sleep_notify)
    remove_did_sleep_notify_listener = partialmethod(Toy._remove_listener, Power.did_sleep_notify)
    enable_battery_voltage_state_change_notify = Power.enable_battery_voltage_state_change_notify
    add_battery_voltage_state_change_notify_listener = partialmethod(Toy._add_listener,
                                                                     Power.battery_voltage_state_change_notify)
    remove_battery_voltage_state_change_notify_listener = partialmethod(Toy._remove_listener,
                                                                        Power.battery_voltage_state_change_notify)

    # SystemInfo
    get_main_app_version = SystemInfo.get_main_app_version
    get_bootloader_version = SystemInfo.get_bootloader_version
    get_board_revision = SystemInfo.get_board_revision
    get_mac_address = SystemInfo.get_mac_address
    get_stats_id = SystemInfo.get_stats_id
    get_secondary_main_app_version = SystemInfo.get_secondary_main_app_version
    add_secondary_main_app_version_notify_listener = partialmethod(Toy._add_listener,
                                                                   SystemInfo.secondary_main_app_version_notify)
    remove_secondary_main_app_version_notify_listener = partialmethod(Toy._remove_listener,
                                                                      SystemInfo.secondary_main_app_version_notify)
    get_processor_name = SystemInfo.get_processor_name
    get_secondary_mcu_bootloader_version = SystemInfo.get_secondary_mcu_bootloader_version
    add_get_secondary_mcu_bootloader_version_notify_listener = partialmethod(Toy._add_listener,
                                                                             SystemInfo.secondary_mcu_bootloader_version_notify)
    remove_get_secondary_mcu_bootloader_version_notify_listener = partialmethod(Toy._remove_listener,
                                                                                SystemInfo.secondary_mcu_bootloader_version_notify)
    get_three_character_sku = SystemInfo.get_three_character_sku

    # Controls
    @property
    @lru_cache(None)
    def drive_control(self):
        return DriveControl(self)

    @property
    @lru_cache(None)
    def firmware_update_control(self):
        return FirmwareUpdateControl(self)

    @property
    @lru_cache(None)
    def multi_led_control(self):
        return LedControl(self)

    @property
    @lru_cache(None)
    def sensor_control(self):
        return SensorControl(self)

    @property
    @lru_cache(None)
    def stats_control(self):
        return StatsControl(self)



import struct
from enum import IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int
from spherov2.listeners.system_info import Version, LastErrorInfo, ConfigBlock, ManufacturingDate, EventLogStatus


class ApplicationIndexes(IntEnum):
    BOOTLOADER = 0
    MAIN_APPLICATION = 1


class ConfigBlockWriteCodes(IntEnum):
    CS_SUCCESS = 0
    CB_BUSY = 1
    BC_FS_ERROR = 2


class DeviceModes(IntEnum):
    FACTORY_MODE = 0
    USER_PLAY_TEST_MODE = 1
    USER_OUT_OF_BOX_MODE = 2
    USER_FULL_MODE = 3


class SosMessages(IntEnum):
    UNKNOWN = 0
    SUBPROCESSOR_CRASHED = 1


class BootReasons(IntEnum):
    COLD_BOOT = 0
    UNEXPECTED_RESET = 1
    APPLICATION_RESET_DUE_TO_ERROR = 2
    APPLICATION_RESET_FOR_A_FIRMWARE_UPDATE = 3
    PROCESSOR_IS_BOOTING_FROM_SLEEP = 4
    PROCESSOR_IS_RESETTING_FOR_SOME_NON_ERROR_REASON = 5


class SystemInfo(Commands):
    _did = 17

    @staticmethod
    def get_main_app_version(toy, proc=None):
        return Version(*struct.unpack('>3H', toy._execute(SystemInfo._encode(toy, 0, proc)).data))

    @staticmethod
    def get_bootloader_version(toy, proc=None):
        return Version(*struct.unpack('>3H', toy._execute(SystemInfo._encode(toy, 1, proc)).data))

    @staticmethod
    def get_board_revision(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 3, proc)).data[0]

    @staticmethod
    def get_mac_address(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 6, proc)).data

    @staticmethod
    def get_model_number(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 18, proc)).data

    @staticmethod
    def get_stats_id(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 19, proc)).data

    @staticmethod
    def get_secondary_main_app_version(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 23, proc))
        return Version(
            *struct.unpack('>3H', toy._wait_packet(SystemInfo.secondary_main_app_version_notify).data))

    secondary_main_app_version_notify = (17, 24, 0xff)

    @staticmethod
    def get_processor_name(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 31, proc)).data.rstrip(b'\0')

    @staticmethod
    def get_boot_reason(toy, proc=None):
        return BootReasons(toy._execute(SystemInfo._encode(toy, 32, proc)).data[0])

    @staticmethod
    def get_last_error_info(toy, proc=None):
        return LastErrorInfo(
            *struct.unpack('>32sH12s', toy._execute(SystemInfo._encode(toy, 33, proc)).data))

    @staticmethod
    def get_secondary_mcu_bootloader_version(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 36, proc))
        return Version(*struct.unpack('>3H', toy._wait_packet(SystemInfo.secondary_mcu_bootloader_version_notify).data))

    secondary_mcu_bootloader_version_notify = (17, 37, 0xff)

    @staticmethod
    def get_three_character_sku(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 40, proc)).data

    @staticmethod
    def write_config_block(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 43, proc))

    @staticmethod
    def get_config_block(toy, proc=None):
        data = toy._execute(SystemInfo.get_config_block(SystemInfo._encode(toy, 44, proc).data))
        return ConfigBlock(*struct.unpack('>2I', data[:8]), data[8:])

    @staticmethod
    def set_config_block(toy, metadata_version, config_block_version, application_data, proc=None):
        toy._execute(SystemInfo._encode(
            45, proc, [*to_bytes(metadata_version, 4), *to_bytes(config_block_version, 4), *application_data]))

    @staticmethod
    def erase_config_block(toy, j, proc=None):
        toy._execute(SystemInfo._encode(toy, 46, proc, to_bytes(j, 4)))

    @staticmethod
    def get_swd_locking_status(toy, proc=None):
        return bool(toy._execute(SystemInfo._encode(toy, 48, proc)).data[0])

    @staticmethod
    def get_manufacturing_date(toy, proc=None):
        return ManufacturingDate(
            *struct.unpack('>HBB', toy._execute(SystemInfo._encode(toy, 51, proc)).data))

    @staticmethod
    def get_sku(toy, proc=None):
        return toy._execute(SystemInfo._encode(toy, 56, proc)).data.rstrip(b'\0')

    @staticmethod
    def get_core_up_time_in_milliseconds(toy, proc=None):
        return to_int(toy._execute(SystemInfo._encode(toy, 57, proc)).data)

    @staticmethod
    def get_event_log_status(toy, proc=None):
        return EventLogStatus(*struct.unpack('>3I', toy._execute(SystemInfo._encode(toy, 58, proc)).data))

    @staticmethod
    def get_event_log_data(toy, j, j2, proc=None):  # unknown name
        return toy._execute(SystemInfo._encode(toy, 59, proc, to_bytes(j, 4) + to_bytes(j2, 4))).data

    @staticmethod
    def clear_event_log(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 60, proc))

    @staticmethod
    def enable_sos_message_notify(toy, enable: bool, proc=None):
        toy._execute(SystemInfo._encode(toy, 61, proc, [int(enable)]))

    sos_message_notify = (17, 62, 0xff), lambda listener, p: listener(SosMessages(p.data[0]))

    @staticmethod
    def get_sos_message(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 63, proc))

    @staticmethod
    def clear_sos_message(toy, proc=None):
        toy._execute(SystemInfo._encode(toy, 68, proc))



from spherov2.commands.sphero import Sphero as SpheroCmd
from spherov2.toy.sphero import Sphero
from spherov2.types import ToyType


class Ollie(Sphero):
    toy_type = ToyType('Ollie', '2B-', '2B', .06)

    get_sku = SpheroCmd.get_sku



import struct
from enum import IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_bytes


class R2DoLegActions(IntEnum):
    UNKNOWN = 0
    THREE_LEGS = 1
    TWO_LEGS = 2
    WADDLE = 3
    TRANSITIONING = 4


class R2LegActions(IntEnum):
    STOP = 0
    THREE_LEGS = 1
    TWO_LEGS = 2
    WADDLE = 3


class SuspensionAnimationModes(IntEnum):
    NON_ACTIVE = 0
    ACTIVE = 1


class Animatronic(Commands):
    _did = 23

    @staticmethod
    def play_animation(toy, animation: IntEnum, proc=None):
        toy._execute(Animatronic._encode(toy, 5, proc, to_bytes(animation, 2)))

    @staticmethod
    def perform_leg_action(toy, leg_action: R2LegActions, proc=None):
        toy._execute(Animatronic._encode(toy, 13, proc, [leg_action]))

    @staticmethod
    def set_head_position(toy, head_position: float, proc=None):
        toy._execute(Animatronic._encode(toy, 15, proc, struct.pack('>f', head_position)))

    play_animation_complete_notify = (23, 17, 0xff)

    @staticmethod
    def get_head_position(toy, proc=None):
        return struct.unpack('>f', toy._execute(Animatronic._encode(toy, 20, proc)).data)[0]

    @staticmethod
    def set_leg_position(toy, leg_position: float, proc=None):
        toy._execute(Animatronic._encode(toy, 21, proc, struct.pack('>f', leg_position)))

    @staticmethod
    def get_leg_position(toy, proc=None):
        return struct.unpack('>f', toy._execute(Animatronic._encode(toy, 22, proc)).data)[0]

    @staticmethod
    def get_leg_action(toy, proc=None):
        return R2DoLegActions(toy._execute(Animatronic._encode(toy, 37, proc)).data[0])

    leg_action_complete_notify = (23, 38, 0xff)

    @staticmethod
    def enable_leg_action_notify(toy, enable: bool, proc=None):
        toy._execute(Animatronic._encode(toy, 42, proc, [int(enable)]))

    @staticmethod
    def stop_animation(toy, proc=None):
        toy._execute(Animatronic._encode(toy, 43, proc))

    @staticmethod
    def enable_idle_animations(toy, enable: bool, proc=None):
        toy._execute(Animatronic._encode(toy, 44, proc, [int(enable)]))

    @staticmethod
    def enable_trophy_mode(toy, enable: bool, proc=None):
        toy._execute(Animatronic._encode(toy, 45, proc, [int(enable)]))

    @staticmethod
    def get_trophy_mode_enabled(toy, proc=None):
        return bool(toy._execute(Animatronic._encode(toy, 46, proc)).data[0])

    @staticmethod
    def enable_head_reset_to_zero_notify(toy, enable: bool, proc=None):
        toy._execute(Animatronic._encode(toy, 57, proc, [int(enable)]))

    head_reset_to_zero_notify = (23, 58, 0xff), lambda listener, _: listener()



from collections import OrderedDict
from functools import partialmethod, lru_cache

from spherov2.commands.async_ import Async
from spherov2.commands.bootloader import Bootloader
from spherov2.commands.core import Core
from spherov2.commands.sphero import Sphero as SpheroCmd
from spherov2.controls.v1 import DriveControl, SensorControl, StatsControl, FirmwareUpdateControl
from spherov2.toy import ToySensor, Toy
from spherov2.types import ToyType


class Sphero(Toy):
    toy_type = ToyType('SPRK/2.0', None, 'Sphero', .06)

    sensors = OrderedDict(
        attitude=OrderedDict(
            pitch=ToySensor(0x40000, -179., 180.),
            roll=ToySensor(0x20000, -179., 180.),
            yaw=ToySensor(0x10000, -179., 180.)
        ),
        accelerometer=OrderedDict(
            x=ToySensor(0x8000, -32768., 32767., lambda x: x / 4096),
            y=ToySensor(0x4000, -32768., 32767., lambda x: x / 4096),
            z=ToySensor(0x2000, -32768., 32767., lambda x: x / 4096)
        ),
        gyroscope=OrderedDict(
            x=ToySensor(0x1000, -20000., 20000., lambda x: x * .1),
            y=ToySensor(0x800, -20000., 20000., lambda x: x * .1),
            z=ToySensor(0x400, -20000., 20000., lambda x: x * .1)
        ),
        back_emf=OrderedDict(
            left=ToySensor(0x40, -32768., 32767.),
            right=ToySensor(0x20, -32768., 32767.),
        )
    )

    extended_sensors = OrderedDict(
        quaternion=OrderedDict(
            x=ToySensor(0x80000000, -10000., 10000., lambda x: x / 10000),
            y=ToySensor(0x40000000, -10000., 10000., lambda x: x / 10000),
            z=ToySensor(0x20000000, -10000., 10000., lambda x: x / 10000),
            w=ToySensor(0x10000000, -10000., 10000., lambda x: x / 10000)
        ),
        locator=OrderedDict(
            x=ToySensor(0x8000000, -32768., 32767.),
            y=ToySensor(0x4000000, -32768., 32767.),
        ),
        accel_one=OrderedDict(accel_one=ToySensor(0x2000000, 0., 8000.)),
        velocity=OrderedDict(
            x=ToySensor(0x1000000, -32768., 32767., lambda x: x * .1),
            y=ToySensor(0x800000, -32768., 32767., lambda x: x * .1),
        ),
        speed=OrderedDict(speed=ToySensor(0x400000, 0., 32767.)),
    )

    def wake(self):
        self._Toy__adapter.write('22bb746f-2bbf-7554-2d6f-726568705327', bytearray([1]))

    # Async
    add_battery_state_changed_notify_listener = partialmethod(Toy._add_listener,
                                                              Async.battery_state_changed_notify)
    remove_battery_state_changed_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Async.battery_state_changed_notify)
    add_collision_detected_notify_listener = partialmethod(Toy._add_listener,
                                                           Async.collision_detected_notify)
    remove_collision_detected_notify_listener = partialmethod(Toy._remove_listener,
                                                              Async.collision_detected_notify)
    add_did_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                  Async.did_sleep_notify)
    remove_did_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                     Async.did_sleep_notify)
    add_gyro_max_notify_listener = partialmethod(Toy._add_listener,
                                                 Async.gyro_max_notify)
    remove_gyro_max_notify_listener = partialmethod(Toy._remove_listener,
                                                    Async.gyro_max_notify)
    add_sensor_streaming_data_notify_listener = partialmethod(Toy._add_listener,
                                                              Async.sensor_streaming_data_notify)
    remove_sensor_streaming_data_notify_listener = partialmethod(Toy._remove_listener,
                                                                 Async.sensor_streaming_data_notify)
    add_will_sleep_notify_listener = partialmethod(Toy._add_listener,
                                                   Async.will_sleep_notify)
    remove_will_sleep_notify_listener = partialmethod(Toy._remove_listener,
                                                      Async.will_sleep_notify)

    # Bootloader
    begin_reflash = Bootloader.begin_reflash
    here_is_page = Bootloader.here_is_page
    jump_to_main = Bootloader.jump_to_main

    # Core
    enable_battery_state_changed_notify = Core.enable_battery_state_changed_notify
    get_bluetooth_info = Core.get_bluetooth_info
    get_charger_state = Core.get_charger_state
    get_power_state = Core.get_power_state
    get_versions = Core.get_versions
    jump_to_bootloader = Core.jump_to_bootloader
    ping = Core.ping
    set_bluetooth_name = Core.set_bluetooth_name
    set_inactivity_timeout = Core.set_inactivity_timeout
    sleep = Core.sleep

    # Sphero
    boost = SpheroCmd.boost
    configure_collision_detection = SpheroCmd.configure_collision_detection
    configure_locator = SpheroCmd.configure_locator
    get_chassis_id = SpheroCmd.get_chassis_id
    get_persistent_options = SpheroCmd.get_persistent_options
    get_temperature = SpheroCmd.get_temperature
    set_temporary_options = SpheroCmd.set_temporary_options
    roll = SpheroCmd.roll
    self_level = SpheroCmd.self_level
    set_back_led_brightness = SpheroCmd.set_back_led_brightness
    set_data_streaming = SpheroCmd.set_data_streaming
    set_heading = SpheroCmd.set_heading
    set_main_led = SpheroCmd.set_main_led
    set_motion_timeout = SpheroCmd.set_motion_timeout
    set_persistent_options = SpheroCmd.set_persistent_options
    set_raw_motors = SpheroCmd.set_raw_motors
    set_rotation_rate = SpheroCmd.set_rotation_rate
    set_stabilization = SpheroCmd.set_stabilization

    # Controls - V1
    @property
    @lru_cache(None)
    def drive_control(self):
        return DriveControl(self)

    # The way utils.py is implemented this doesn't need to be present
    # @property
    # @lru_cache(None)
    # def multi_led_control(self):
    #    return LedControl(self)

    @property
    @lru_cache(None)
    def sensor_control(self):
        return SensorControl(self)

    @property
    @lru_cache(None)
    def stats_control(self):
        return StatsControl(self)

    @property
    @lru_cache(None)
    def firmware_update_control(self):
        return FirmwareUpdateControl(self)



import struct
from enum import IntEnum, IntFlag

from spherov2.listeners.async_ import CollisionDetected
from spherov2.listeners.core import PowerStates


class CollisionAxis(IntFlag):
    X_AXIS = 0x1  # 0b1
    Y_AXIS = 0x2  # 0b10


class GyroMaxExceedsFlags(IntFlag):
    X_POSITIVE = 0x1  # 0b1
    X_NEGATIVE = 0x2  # 0b10
    Y_POSITIVE = 0x4  # 0b100
    Y_NEGATIVE = 0x8  # 0b1000
    Z_POSITIVE = 0x10  # 0b10000
    Z_NEGATIVE = 0x20  # 0b100000


class PowerStates(IntEnum):
    UNKNOWN = 0
    CHARGING = 1
    OK = 2
    LOW = 3
    CRITICAL = 4


class Async:
    battery_state_changed_notify = (0xfe, 1), lambda listener, p: listener(PowerStates(p.data[0]))
    sensor_streaming_data_notify = (
        (0xfe, 3), lambda listener, p: listener(list(struct.unpack('>%dh' % (len(p.data) // 2), p.data))))
    will_sleep_notify = (0xfe, 5), lambda listener, _: listener()

    @staticmethod
    def __process(listener, packet):
        unpacked = struct.unpack('>3hB2hBL', packet.data)
        listener(CollisionDetected(acceleration_x=unpacked[0] / 4096, acceleration_y=unpacked[1] / 4096,
                                   acceleration_z=unpacked[2] / 4096, x_axis=bool(unpacked[3] & 1),
                                   y_axis=bool(unpacked[3] & 2), power_x=unpacked[4], power_y=unpacked[5],
                                   speed=unpacked[6], time=unpacked[7] / 1000))

    collision_detected_notify = (0xfe, 7), __process.__func__
    gyro_max_notify = (0xfe, 12), lambda listener, p: listener(p.data[0])
    did_sleep_notify = (0xfe, 20), lambda listener, _: listener()



from enum import IntFlag, IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_int


class PendingUpdateFlags(IntFlag):
    NORDIC = 0x1  # 0b1
    ST = 0x2  # 0b10
    ESP2886 = 0x4  # 0b100
    AUDIO = 0x8  # 0b1000
    ANIMATIONS = 0x10  # 0b10000
    ST_BOOTLOADER = 0x20  # 0b100000


class ApplicationIds(IntEnum):
    BOOTLOADER = 0
    MAIN_APP = 1


class MainAppValidaties(IntEnum):
    UNKNOWN = 0
    UNKNOWN_BECAUSE_UPDATE_FLAG_SET = 1
    VALID = 2
    INVALID = 3


class ResetStrategies(IntEnum):
    RESET_INTO_OR_JUMP_TO_MAIN_APP = 1
    RESET_INTO_OR_JUMP_TO_BOOTLOADER = 2


class UpdateMethods(IntEnum):
    REQUIRES_RESET = 0
    MASTER_CONTROLS_UPDATE_FLOW = 1
    IN_MASTER_BOOTLOADER = 2


class Firmware(Commands):
    _did = 29

    @staticmethod
    def get_pending_update_flags(toy, proc=None):
        return PendingUpdateFlags(to_int(toy._execute(Firmware._encode(toy, 13, proc)).data))

    @staticmethod
    def get_current_application_id(toy, proc=None):
        return ApplicationIds(toy._execute(Firmware._encode(toy, 21, proc)).data[0])

    @staticmethod
    def get_all_updatable_processors(toy, proc=None):
        toy._execute(Firmware._encode(toy, 22, proc))

    @staticmethod
    def get_version_for_updatable_processors(toy, proc=None):
        toy._execute(Firmware._encode(toy, 24, proc))

    @staticmethod
    def set_pending_update_for_processors(toy, data, proc=None):  # unknown names
        return ResetStrategies(toy._execute(Firmware._encode(toy, 26, proc, data)).data[0])

    @staticmethod
    def get_pending_update_for_processors(toy, proc=None):
        return toy._execute(Firmware._encode(toy, 27, proc)).data

    @staticmethod
    def reset_with_parameters(toy, strategy, proc=None):
        toy._execute(Firmware._encode(toy, 28, proc, [strategy]))

    @staticmethod
    def clear_pending_update_processors(toy, data, proc=None):  # unknown names
        toy._execute(Firmware._encode(toy, 38, proc, data))



from enum import IntEnum
from typing import NamedTuple


class Versions(NamedTuple):
    record_version: int
    model_number: int
    hardware_version_code: int
    main_app_version_major: int
    main_app_version_minor: int
    bootloader_version: str
    orb_basic_version: str
    overlay_version: str


class BluetoothInfo(NamedTuple):
    name: bytes
    address: bytes


class PowerStates(IntEnum):
    UNKNOWN = 0
    CHARGING = 1
    OK = 2
    LOW = 3
    CRITICAL = 4


class PowerState(NamedTuple):
    record_version: int
    state: PowerStates
    voltage: float
    number_of_charges: int
    time_since_last_charge: int



from enum import IntEnum

from spherov2.commands import Commands


class BleCentralStates(IntEnum):
    NOT_YET_INITIALIZED = 0
    DISCONNECTED = 1
    SCANNING = 2
    CONNECTING = 3
    RECONNECTING = 4
    CONNECTED = 5
    DISCONNECTING = 6
    CONFIGURING = 7


class BleCentralStates(IntEnum):
    NOT_YET_INITIALIZED = 0
    DISCONNECTED = 1
    SCANNING = 2
    CONNECTING = 3
    RECONNECTING = 4
    CONNECTED = 5
    DISCONNECTING = 6
    CONFIGURING = 7


class BleCentralStates(IntEnum):
    NOT_YET_INITIALIZED = 0
    DISCONNECTED = 1
    SCANNING = 2
    CONNECTING = 3
    RECONNECTING = 4
    CONNECTED = 5
    DISCONNECTING = 6
    CONFIGURING = 7


class Connection(Commands):
    _did = 25

    @staticmethod
    def set_bluetooth_name(toy, name: bytes, proc=None):
        toy._execute(Connection._encode(toy, 3, proc, [*name, 0]))

    @staticmethod
    def get_bluetooth_name(toy, proc=None):
        return toy._execute(Connection._encode(toy, 4, proc)).data.rstrip(b'\0')

    @staticmethod
    def get_bluetooth_advertising_name(toy, proc=None):
        return toy._execute(Connection._encode(toy, 5, proc)).data.rstrip(b'\0')



import struct
from enum import IntFlag, IntEnum

from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int


class AudioPlaybackModes(IntFlag):
    PLAY_IMMEDIATELY = 0x0  # 0b0
    PLAY_ONLY_IF_NOT_PLAYING = 0x1  # 0b1
    PLAY_AFTER_CURRENT_SOUND = 0x2  # 0b10


class UsbConnectionStatus(IntEnum):
    UNKNOWN = 0
    CONNECTED_BUT_NOT_READY = 1
    CONNECTED_AND_READY = 2
    DISCONNECTED = 3


class FadeOverrideOptions(IntEnum):
    NONE = 0
    NO_FADING = 1
    FADING_ANIMATIONS = 2


class SpecdrumsColorPaletteIndicies(IntEnum):
    DEFAULT = 0
    MIDI = 1


class FrameInfoTypes(IntEnum):
    COMPRESSED_FRAME_PLAYER_INFO_TYPE_8_BIT = 0
    COMPRESSED_FRAME_PLAYER_INFO_TYPE_16_BIT = 1
    COMPRESSED_FRAME_PLAYER_INFO_TYPE_32_BIT = 2
    COMPRESSED_FRAME_PLAYER_INFO_TYPE_64_BIT = 3


class FrameRotationOptions(IntEnum):
    NORMAL = 0
    ROTATE_90_DEGREES = 1
    ROTATE_180_DEGREES = 2
    ROTATE_270_DEGREES = 3


class TextScrollingReasonCodes(IntEnum):
    DONE = 0
    LOOPING = 1


class IO(Commands):
    _did = 26

    @staticmethod
    def set_led(toy, s, s2, s3, s4, b, proc=None):  # Untested / Unknown Param Names
        toy._execute(IO._encode(toy, 4, proc, [s, s2, s3, s4, b]))

    @staticmethod
    def play_audio_file(toy, sound, playback_mode: AudioPlaybackModes, proc=None):
        toy._execute(IO._encode(toy, 7, proc, [*to_bytes(sound, 2), playback_mode]))

    @staticmethod
    def set_audio_volume(toy, volume, proc=None):
        toy._execute(IO._encode(toy, 8, proc, [volume]))

    @staticmethod
    def get_audio_volume(toy, proc=None):
        return toy._execute(IO._encode(toy, 9, proc)).data[0]

    @staticmethod
    def stop_all_audio(toy, proc=None):
        toy._execute(IO._encode(toy, 10, proc))

    @staticmethod
    def set_all_leds_with_16_bit_mask(toy, mask, values, proc=None):
        toy._execute(IO._encode(toy, 14, proc, [*to_bytes(mask, 2), *values]))

    @staticmethod
    def start_idle_led_animation(toy, proc=None):
        toy._execute(IO._encode(toy, 25, proc))

    @staticmethod
    def set_all_leds_with_32_bit_mask(toy, mask, values, proc=None):
        toy._execute(IO._encode(toy, 26, proc, [*to_bytes(mask, 4), *values]))

    @staticmethod
    def set_all_leds_with_8_bit_mask(toy, mask, values, proc=None):  # Untested / Unknown Param Names
        toy._execute(IO._encode(toy, 28, proc, [*to_bytes(mask, 1), *values]))

    @staticmethod
    def enable_color_tap_notify(toy, z, proc=None):  # Untested / Unknown Param Names
        toy._execute(IO._encode(toy, 29, proc, [z]))

    color_tap_notify = (26, 30, 0xff), lambda listener, p: listener(p.data[0])  # Untested / Unknown Param Names

    @staticmethod
    def set_compressed_frame_player_pixel(toy, x, y, r, g, b, proc=None):
        toy._execute(IO._encode(toy, 45, proc, [x, y, r, g, b]))

    @staticmethod
    def set_compressed_frame_player(toy, s_arr, proc=None):
        toy._execute(IO._encode(toy, 46, proc, [*s_arr]))

    @staticmethod
    def set_compressed_frame_player_one_color(toy, r, g, b, proc=None):
        toy._execute(IO._encode(toy, 47, proc, [r, g, b]))

    @staticmethod
    def save_compressed_frame_player64_bit_frame(toy, frame_index, compressed_frame, proc=None):
        toy._execute(IO._encode(toy, 48, proc, [*to_bytes(frame_index, 2), *compressed_frame]))

    @staticmethod
    def save_compressed_frame_player_animation(toy, animation_id, fps: int, fade_animation:bool, palette_colors, frames_indexes, proc=None):
        toy._execute(IO._encode(
            toy, 49, proc, [animation_id, fps % 31, int(fade_animation), len(palette_colors) // 3, *palette_colors, *struct.pack('>%dH' % (len(frames_indexes) + 1), len(frames_indexes), *frames_indexes)]))

    @staticmethod
    def play_compressed_frame_player_animation(toy, animation_id, proc=None):
        toy._execute(IO._encode(toy, 50, proc, [animation_id]))

    @staticmethod
    def play_compressed_frame_player_frame(toy, i, proc=None):  # unknown names
        toy._execute(IO._encode(toy, 51, proc, to_bytes(i, 2)))

    @staticmethod
    def get_compressed_frame_player_list_of_frames(toy, proc=None):
        data = toy._execute(IO._encode(toy, 52, proc)).data
        return struct.unpack('>%dH' % (len(data) // 2), data)

    @staticmethod
    def delete_all_compressed_frame_player_animations_and_frames(toy, proc=None):
        toy._execute(IO._encode(toy, 53, proc))

    @staticmethod
    def pause_compressed_frame_player_animation(toy, proc=None):
        toy._execute(IO._encode(toy, 54, proc))

    @staticmethod
    def resume_compressed_frame_player_animation(toy, proc=None):
        toy._execute(IO._encode(toy, 55, proc))

    @staticmethod
    def reset_compressed_frame_player_animation(toy, proc=None):
        toy._execute(IO._encode(toy, 56, proc))

    @staticmethod
    def override_compressed_frame_player_animation_global_settings(toy, fps:int, fade_options:FadeOverrideOptions, proc=None):  # Untested / Unknown Param Names
        toy._execute(IO._encode(toy, 57, proc, [fps, fade_options]))

    @staticmethod
    def set_compressed_frame_player_frame_rotation(toy, rotation: FrameRotationOptions, proc=None):
        toy._execute(IO._encode(toy, 58, proc, [rotation.value]))

    @staticmethod
    def set_compressed_frame_player_text_scrolling(toy, str_to_display: str, r, g, b, speed: int, repeat: bool, proc=None):
        toy._execute(IO._encode(toy, 59, proc, [r, g, b, speed % 0x1f, int(repeat), *[ord(c) for c in str_to_display[:25]], 0x00]))

    set_compressed_frame_player_text_scrolling_notify = (26, 60, 0xff), lambda listener, p: listener(
        p.data[0])  # Untested / Unknown Param Names

    @staticmethod
    def draw_compressed_frame_player_line(toy, x1, y1, x2, y2, r, g, b, proc=None):
        toy._execute(IO._encode(toy, 61, proc, [x1, y1, x2, y2, r, g, b]))

    @staticmethod
    def draw_compressed_frame_player_fill(toy, x1, y1, x2, y2, r, g, b, proc=None):
        toy._execute(IO._encode(toy, 62, proc, [x1, y1, x2, y2, r, g, b]))

    compressed_frame_player_animation_complete_notify = (26, 63, 0xff), lambda listener, p: listener(p.data[0])

    @staticmethod
    def assign_compressed_frame_player_frames_to_animation(toy, s, i, i_arr, proc=None):  # unknown names
        toy._execute(IO._encode(toy, 64, proc, [s, *struct.pack('>%dH' % (len(i_arr) + 1), i, *i_arr)]))

    @staticmethod
    def save_compressed_frame_player_animation_without_frames(toy, s, s2, z, s3, s_arr, i, proc=None):  # unknown names
        toy._execute(IO._encode(toy, 65, proc, [s, s2, int(z), s3, *s_arr, *to_bytes(i, 2)]))

    @staticmethod
    def set_compressed_frame_player_single_character(toy, r:int, g:int, b:int, character:str, proc=None):
        toy._execute(IO._encode(toy, 66, proc, [r, g, b, ord(character)]))

    @staticmethod
    def play_compressed_frame_player_animation_with_loop_option(toy, animation_id, loop, proc=None):
        toy._execute(IO._encode(toy, 67, proc, [animation_id, int(loop)]))

    @staticmethod
    def get_active_color_palette(toy, proc=None):
        return toy._execute(IO._encode(toy, 68, proc)).data

    @staticmethod
    def set_active_color_palette(toy, rgb_index_bytes, proc=None):
        toy._execute(IO._encode(toy, 69, proc, rgb_index_bytes))

    @staticmethod
    def get_color_identification_report(toy, red, green, blue, confidence_threshold, proc=None):
        return toy._execute(IO._encode(toy, 70, proc, [red, green, blue, confidence_threshold])).data

    @staticmethod
    def load_color_palette(toy, palette_index, proc=None):
        toy._execute(IO._encode(toy, 71, proc, [palette_index]))

    @staticmethod
    def save_color_palette(toy, palette_index, proc=None):
        toy._execute(IO._encode(toy, 72, proc, [palette_index]))

    @staticmethod
    def get_compressed_frame_player_frame_info_type(toy, proc=None):
        return FrameInfoTypes(toy._execute(IO._encode(toy, 76, proc)).data[0])

    @staticmethod
    def save_compressed_frame_player16_bit_frame(toy, i, i2, i3, i4, i5, proc=None):  # unknown names
        return to_int(toy._execute(IO._encode(toy, 77, proc, struct.pack('>5H', i, i2, i3, i4, i5))).data)

    @staticmethod
    def release_led_requests(toy, proc=None):
        toy._execute(IO._encode(toy, 78, proc))



from enum import IntEnum, IntFlag

from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int
from spherov2.listeners.sphero import Options


class ConfigurationOptions(IntFlag):
    DISABLE_SLEEP_IN_CHARGE = 0x1  # 0b1
    ENABLE_VECTOR_DRIVE = 0x2  # 0b10
    DISABLE_SELF_LEVEL_IN_CHARGER = 0x4  # 0b100
    ENABLE_TAIL_LIGHT_ALWAYS_ON = 0x8  # 0b1000
    ENABLE_MOTION_TIMEOUT = 0x10  # 0b10000
    ENABLE_GYRO_MAX_NOTIFY = 0x100  # 0b100000000
    ENABLE_FULL_SPEED = 0x400  # 0b10000000000


class SelfLevelOptions(IntFlag):
    START = 0x1  # 0b1
    KEEP_HEADING = 0x2  # 0b10
    SLEEP_AFTER = 0x4  # 0b100
    TURN_CONTROL_SYSTEM_ON = 0x8  # 0b1000


class CollisionDetectionMethods(IntEnum):
    OFF = 0
    DEFAULT = 1


class DeviceModes(IntEnum):
    NORMAL = 0
    HACK = 1


class RollModes(IntEnum):
    STOP = 0
    GO = 1
    CALIBRATE = 2


class ReverseFlags(IntEnum):
    OFF = 0
    ON = 1


class RawMotorModes(IntEnum):
    OFF = 0
    FORWARD = 1
    REVERSE = 2
    BRAKE = 3
    IGNORE = 4


class Sphero(Commands):
    _did = 2

    @staticmethod
    def set_heading(toy, heading: int, proc=None):
        toy._execute(Sphero._encode(toy, 1, proc, to_bytes(heading, 2)))

    @staticmethod
    def set_stabilization(toy, stabilize: bool, proc=None):
        toy._execute(Sphero._encode(toy, 2, proc, [int(stabilize)]))

    @staticmethod
    def set_rotation_rate(toy, rate: int, proc=None):
        toy._execute(Sphero._encode(toy, 3, proc, [rate]))

    @staticmethod
    def get_chassis_id(toy, proc=None):
        return to_int(toy._execute(Sphero._encode(toy, 7, proc)).data)

    @staticmethod
    def self_level(toy, opt1: bool, opt2: bool, opt3: bool, opt4: bool, angle_limit, timeout, true_time, proc=None):
        # unknown names
        toy._execute(Sphero._encode(
            toy, 9, proc, [opt1 | (opt2 << 1) | (opt3 << 2) | (opt4 << 3), angle_limit, timeout, true_time]))

    @staticmethod
    def set_data_streaming(toy, interval, num_samples_per_packet, mask, count, extended_mask, proc=None):
        toy._execute(Sphero._encode(
            toy, 17, proc,
            [*to_bytes(interval, 2), *to_bytes(num_samples_per_packet, 2), *to_bytes(mask, 4), count & 0xff,
             *to_bytes(extended_mask, 4)]))

    @staticmethod
    def configure_collision_detection(toy, collision_detection_method: CollisionDetectionMethods,
                                      x_threshold, y_threshold, x_speed, y_speed, dead_time, proc=None):
        toy._execute(Sphero._encode(
            toy, 18, proc, [collision_detection_method, x_threshold, y_threshold, x_speed, y_speed, dead_time]))

    @staticmethod
    def configure_locator(toy, flags, x, y, yaw_tare, proc=None):
        toy._execute(Sphero._encode(toy, 19, proc, [flags, *to_bytes(x, 2), *to_bytes(y, 2), *to_bytes(yaw_tare, 2)]))

    @staticmethod
    def get_temperature(toy, proc=None):
        data = toy._execute(Sphero._encode(toy, 22, proc)).data
        return data[0] + data[1] / 10

    @staticmethod
    def set_main_led(toy, r, g, b, proc=None):
        toy._execute(Sphero._encode(toy, 32, proc, [r, g, b]))

    @staticmethod
    def set_back_led_brightness(toy, brightness, proc=None):
        toy._execute(Sphero._encode(toy, 33, proc, data=[brightness]))

    @staticmethod
    def roll(toy, speed, heading, roll_mode: RollModes, reverse_flag: ReverseFlags, proc=None):
        toy._execute(Sphero._encode(toy, 48, proc, [speed, *to_bytes(heading, 2), roll_mode, reverse_flag]))

    @staticmethod
    def boost(toy, s, s2, proc=None):  # unknown names
        toy._execute(Sphero._encode(toy, 49, proc, [s, *to_bytes(s2, 2)]))

    @staticmethod
    def set_raw_motors(toy, left_mode: RawMotorModes, left_speed, right_mode: RawMotorModes, right_speed, proc=None):
        toy._execute(Sphero._encode(toy, 51, proc, [left_mode, left_speed, right_mode, right_speed]))

    @staticmethod
    def set_motion_timeout(toy, timeout: int, proc=None):
        toy._execute(Sphero._encode(toy, 52, proc, to_bytes(timeout, 2)))

    @staticmethod
    def set_persistent_options(toy, options: Options, proc=None):  # unknown names
        toy._execute(Sphero._encode(toy, 53, proc, to_bytes(
            options.disable_sleep_in_charger | (options.enable_vector_drive << 1) | (
                    options.disable_self_level_in_charger << 2) | (options.enable_tail_light_always_on << 3) | (
                    options.enable_motion_timeout << 4) | (options.enable_gyro_max_notify << 8) | (
                    options.enable_full_speed << 10), 4)))

    @staticmethod
    def get_persistent_options(toy, proc=None):
        data = to_int(toy._execute(Sphero._encode(toy, 54, proc)).data)
        return Options(bool(data & 1), bool(data & 2), bool(data & 4), bool(data & 8), bool(data & 16),
                       bool(data & 256), bool(data & 1024))

    @staticmethod
    def set_temporary_options(toy, options: Options, proc=None):
        toy._execute(Sphero._encode(toy, 55, proc, to_bytes(
            options.disable_sleep_in_charger | (options.enable_vector_drive << 1) | (
                    options.disable_self_level_in_charger << 2) | (options.enable_tail_light_always_on << 3) | (
                    options.enable_motion_timeout << 4) | (options.enable_gyro_max_notify << 8) | (
                    options.enable_full_speed << 10), 4)))

    @staticmethod
    def get_temporary_options(toy, proc=None):
        data = to_int(toy._execute(Sphero._encode(toy, 56, proc)).data)
        return Options(bool(data & 1), bool(data & 2), bool(data & 4), bool(data & 8), bool(data & 16),
                       bool(data & 256), bool(data & 1024))

    @staticmethod
    def get_sku(toy, proc=None):
        return reversed(toy._execute(Sphero._encode(toy, 58, proc)).data)



import socket
import struct
import threading
from concurrent import futures
from typing import NamedTuple

from spherov2.adapter.tcp_consts import RequestOp, ResponseOp
from spherov2.helper import to_int, to_bytes


class MockDevice(NamedTuple):
    name: str
    address: str


def recvall(s, size):
    data = bytes()
    while len(data) < size:
        n = s.recv(size - len(data))
        if not n:
            raise EOFError
        data += n
    return data


def get_tcp_adapter(host: str, port: int = 50004):
    """Gets an anonymous ``TCPAdapter`` with the given address and port."""

    class TCPAdapter:
        @staticmethod
        def scan_toys(timeout=5.0):
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, port))
            try:
                s.sendall(RequestOp.SCAN + struct.pack('!f', timeout))
                code = recvall(s, 1)
                if code == ResponseOp.ERROR:
                    size = to_int(recvall(s, 2))
                    data = recvall(s, size)
                    raise Exception(data.decode('utf_8'))
                elif code != ResponseOp.OK:
                    raise SystemError(f'Unexpected response op code {code}')
                num_devices = to_int(recvall(s, 2))
                devices = []
                for _ in range(num_devices):
                    name_size = to_int(recvall(s, 2))
                    name = recvall(s, name_size).decode('utf_8')
                    address_size = to_int(recvall(s, 2))
                    addr = recvall(s, address_size).decode('ascii')
                    devices.append(MockDevice(name, addr))
                return devices
            finally:
                s.sendall(RequestOp.END)
                s.close()

        @staticmethod
        def scan_toy(name: str, timeout: float = 5.0):
            name = name.encode('utf_8')
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((host, port))
            try:
                s.sendall(RequestOp.FIND + to_bytes(len(name), 2) +
                          name + struct.pack('!f', timeout))
                code = recvall(s, 1)
                if code == ResponseOp.ERROR:
                    size = to_int(recvall(s, 2))
                    if size == 0:
                        return None
                    data = recvall(s, size)
                    raise Exception(data.decode('utf_8'))
                elif code != ResponseOp.OK:
                    raise SystemError(f'Unexpected response op code {code}')
                name_size = to_int(recvall(s, 2))
                name = recvall(s, name_size).decode('utf_8')
                address_size = to_int(recvall(s, 2))
                addr = recvall(s, address_size).decode('ascii')
                return MockDevice(name, addr)
            finally:
                s.sendall(RequestOp.END)
                s.close()

        def __init__(self, address):
            self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__socket.connect((host, port))
            address = address.encode('ascii')

            self.__sequence = 0
            self.__sequence_wait = {}

            self.__callbacks = {}
            self.__thread = threading.Thread(target=self.__recv)
            self.__thread.start()
            try:
                self.__send(RequestOp.INIT, to_bytes(
                    len(address), 2) + address)
            except:
                self.close()
                raise

        def __recv(self):
            while True:
                try:
                    code = recvall(self.__socket, 1)
                except:
                    break
                if code == ResponseOp.OK:
                    self.__sequence_wait.pop(
                        recvall(self.__socket, 1)[0]).set_result(None)
                    continue
                size = to_int(recvall(self.__socket, 2))
                data = recvall(self.__socket, size)
                if code == ResponseOp.ON_DATA:
                    uuid = data.decode('ascii').lower()
                    size = recvall(self.__socket, 1)[0]
                    data = recvall(self.__socket, size)
                    for f in self.__callbacks.get(uuid, []):
                        f(uuid, data)
                elif code == ResponseOp.ERROR:
                    err = Exception(data.decode('utf_8'))
                    self.__sequence_wait.pop(recvall(self.__socket, 1)[
                                             0]).set_exception(err)

        def __send(self, cmd, payload):
            if not self.__thread.is_alive():
                raise ConnectionError('Connection is lost')
            seq = self.__sequence
            self.__sequence = (self.__sequence + 1) % 0x100
            f = self.__sequence_wait[seq] = futures.Future()
            self.__socket.sendall(cmd + bytes([seq]) + payload)
            f.result()

        def close(self):
            self.__socket.sendall(RequestOp.END)
            self.__socket.close()
            self.__thread.join()

        def set_callback(self, uuid, cb):
            if uuid in self.__callbacks:
                self.__callbacks[uuid].add(cb)
            else:
                self.__callbacks[uuid] = {cb}
                buf = uuid.encode('ascii')
                self.__send(RequestOp.SET_CALLBACK,
                            to_bytes(len(buf), 2) + buf)

        def write(self, uuid, data):
            uuid = uuid.encode('ascii')
            self.__send(RequestOp.WRITE, to_bytes(len(uuid), 2) +
                        uuid + to_bytes(len(data), 2) + data)

    return TCPAdapter



from typing import NamedTuple, Union


class ToyType(NamedTuple):
    display_name: str
    prefix: Union[str, None]
    filter_prefix: str
    cmd_safe_interval: float


class Color(NamedTuple):
    r: int = None
    g: int = None
    b: int = None



from functools import lru_cache

from spherov2.types import Color


def to_int(ba):
    return int.from_bytes(ba, byteorder='big')


def to_bytes(i: int, size):
    return i.to_bytes(size, byteorder='big')


@lru_cache(None)
def nibble_to_byte(high, low):
    return (high << 4) | low


def bound_value(lower, value, upper):
    return min(upper, max(lower, value))


def bound_color(color: Color, default_color: Color):
    return Color(
        r=default_color.r if color.r is None else bound_value(0, color.r, 255),
        g=default_color.g if color.g is None else bound_value(0, color.g, 255),
        b=default_color.b if color.b is None else bound_value(0, color.b, 255)
    )


def packet_chk(payload):
    return 0xff - (sum(payload) & 0xff)



import struct
from enum import IntEnum, IntFlag

from spherov2.commands import Commands
from spherov2.helper import to_int
from spherov2.listeners.power import BatteryVoltageStateThresholds


class PowerOptions(IntFlag):
    SLEEP_WHILE_CHARGING = 0x1  # 0b1
    DOUBLE_TAP_TO_WAKE = 0x2  # 0b10


class BatteryVoltageAndStateStates(IntEnum):
    CHARGED = 0
    CHARGING = 1
    NOT_CHARGING = 2
    OK = 3
    LOW = 4
    CRITICAL = 5
    RESERVED = 6
    UNUSED = 7


class BatteryVoltageStates(IntEnum):
    UNKNOWN = 0
    OK = 1
    LOW = 2
    CRITICAL = 3


class BatteryStates(IntEnum):
    CHARGED = 0
    CHARGING = 1
    NOT_CHARGING = 2
    OK = 3
    LOW = 4
    CRITICAL = 5
    UNKNOWN = 255


class BatteryVoltageReadingTypes(IntEnum):
    CALIBRATED_AND_FILTERED = 0
    CALIBRATED_AND_UNFILTERED = 1
    UNCALIBRATED_AND_UNFILTERED = 2


class ChargerStates(IntEnum):
    UNKNOWN = 0
    NOT_CHARGING = 1
    CHARGING = 2
    CHARGED = 3


class AmplifierIds(IntEnum):
    LEFT_MOTOR = 0
    RIGHT_MOTOR = 1


class EfuseIds(IntEnum):
    PRIMARY_EFUSE = 0


class Power(Commands):
    _did = 19

    @staticmethod
    def enter_deep_sleep(toy, s, proc=None):
        toy._execute(Power._encode(toy, 0, proc, [s]))

    @staticmethod
    def sleep(toy, proc=None):
        toy._execute(Power._encode(toy, 1, proc))

    @staticmethod
    def get_battery_voltage(toy, proc=None):
        return to_int(toy._execute(Power._encode(toy, 3, proc)).data) / 100

    @staticmethod
    def get_battery_state(toy, proc=None):
        return BatteryStates(toy._execute(Power._encode(toy, 4, proc)).data[0])

    @staticmethod
    def enable_battery_state_changed_notify(toy, enable: bool, proc=None):
        toy._execute(Power._encode(toy, 5, proc, [int(enable)]))

    battery_state_changed_notify = (19, 6, 0xff), lambda listener, p: listener(BatteryVoltageAndStateStates(p.data[0]))

    @staticmethod
    def force_battery_refresh(toy, proc=None):
        toy._execute(Power._encode(toy, 12, proc))

    @staticmethod
    def wake(toy, proc=None):
        toy._execute(Power._encode(toy, 13, proc))

    @staticmethod
    def get_battery_percentage(toy, proc=None):
        return toy._execute(Power._encode(toy, 16, proc)).data[0]

    @staticmethod
    def get_battery_voltage_state(toy, proc=None):
        return BatteryVoltageStates(toy._execute(Power._encode(toy, 23, proc)).data[0])

    will_sleep_notify = (19, 25, 0xff), lambda listener, _: listener()
    did_sleep_notify = (19, 26, 0xff), lambda listener, _: listener()

    @staticmethod
    def enable_battery_voltage_state_change_notify(toy, enable: bool, proc=None):
        toy._execute(Power._encode(toy, 27, proc, [int(enable)]))

    battery_voltage_state_change_notify = (19, 28, 0xff), lambda listener, p: listener(BatteryVoltageStates(p.data[0]))

    @staticmethod
    def get_charger_state(toy, proc=None):  # Untested
        return toy._execute(Power._encode(toy, 31, proc)).data[0]

    @staticmethod
    def enable_charger_state_changed_notify(toy, enable: bool, proc=None):
        toy._execute(Power._encode(toy, 32, proc, [int(enable)]))

    charger_state_changed_notify = (19, 33, 0xff), lambda listener, p: listener(
        BatteryVoltageStates(p.data[0]))  # Untested / Unknown param name

    @staticmethod
    def get_battery_adc_reading(toy, proc=None):  # Untested
        return toy._execute(Power._encode(toy, 34, proc)).data[0]

    @staticmethod
    def set_battery_calibration_slope_and_intercept(toy, f, f2, proc=None):  # untested / Unknown param names
        toy._execute(Power._encode(toy, 35, proc, [f, f2]))

    @staticmethod
    def get_battery_calibration_slope_intercept(toy, proc=None):  # Untested
        return toy._execute(Power._encode(toy, 36, proc)).data[0]

    @staticmethod
    def get_battery_voltage_in_volts(toy, reading_type: BatteryVoltageReadingTypes, proc=None):
        return struct.unpack('>f', toy._execute(Power._encode(toy, 37, proc, [reading_type])).data)[0]

    @staticmethod
    def get_battery_voltage_state_thresholds(toy, proc=None):
        return BatteryVoltageStateThresholds(
            *struct.unpack('>3f', toy._execute(Power._encode(toy, 38, proc)).data))

    @staticmethod
    def get_current_sense_amplifier_current(toy, amplifier_id: AmplifierIds, proc=None):
        return struct.unpack('>f', toy._execute(Power._encode(toy, 39, proc, [amplifier_id])).data)[0]

    @staticmethod
    def get_efuse_fault_status(toy, efuse_id: EfuseIds, proc=None):
        return toy._execute(Power._encode(toy, 40, proc, [efuse_id])).data[0]

    efuse_fault_occurred_notify = (19, 41, 0xff), lambda listener, p: listener(EfuseIds(p.data[0]))

    @staticmethod
    def enable_efuse(toy, efuse_id: EfuseIds, proc=None):
        toy._execute(Power._encode(toy, 42, proc, [efuse_id]))



from setuptools import setup, find_namespace_packages

with open('README.md', 'rt', encoding='utf_8') as f:
    long_desc = f.read()

setup(
    name='spherov2',
    version='0.12.1',
    author='Hanbang Wang', # and 'Elionardo Feliciano',
    author_email='hanbangw@cis.upenn.edu', # and 'elionardo.feliciano.dev@gmail.com',
    license='MIT',
    description='An unofficial Bluetooth low energy library for Sphero toys in Python.',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url='https://github.com/artificial-intelligence-class/spherov2.py',
    packages=find_namespace_packages(include=['spherov2', 'spherov2.*'],
                                     exclude=["*.test", "*.test.*"]),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Robot Framework :: Library',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    keywords='robotics Sphero toy bluetooth ble',
    python_requires='>=3.7',
    install_requires=['numpy', 'transforms3d']
)
